// ----------------------------------------------------------------------------
// Copyright 2017 Nervana Systems Inc.
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// ----------------------------------------------------------------------------

#pragma once

#include "ngraph/node.hpp"
#include "ngraph/types/type.hpp"

namespace ngraph
{
    class Function;
    namespace op
    {
        /// \brief A function parameter.
        ///
        /// Parameters are nodes that represent the arguments that will be passed to user-defined functions.
        /// Function creation requires a sequence of parameters.
        /// Basic graph operations do not need parameters attached to a function.
        ///
        /// ## Parameters
        ///
        /// |              | Description                        |
        /// | -------------| ---------------------------------- |
        /// | `value_type` | The type \f$T\f$ of the parameter. |
        ///
        /// ## Output
        ///
        /// | Type    | Description                                                                                                                 |
        /// | ------- | --------------------------------------------------------------------------------------------------------------------------- |
        /// | \f$T\f$ | The value of the parameter, supplied by the `FunctionCall` to this function or in the initial `ngraph::runtime::CallFrame`. |
        class Parameter : public Node
        {
        protected:
            virtual void generate_adjoints(autodiff::Adjoints& adjoints,
                                           const std::shared_ptr<Node>& delta) override;

        public:
            /// \brief Constructions a parameter node.
            ///
            /// \param value_type The type of the parameter.
            Parameter(const std::shared_ptr<const ValueType>& value_type);
            /// \brief Constructions a tensor view-typed parameter node.
            ///
            /// \param element_type The element type of the parameter.
            /// \param shape The shape of the parameter.
            Parameter(const ngraph::element::Type& element_type, const Shape& shape);

            virtual std::shared_ptr<Node> copy_with_new_args(
                const std::vector<std::shared_ptr<Node>>& new_args) const override
            {
                if (new_args.size() != 0)
                    throw ngraph_error("Incorrect number of new arguments");
                return std::make_shared<Parameter>(get_value_type());
            }
        };
    }
}
